/**
 * All-in-one start script, starts everything you need to try N|Solid Console.
 *
 * 1. Start service dependencies e.g. etcd, nsolid-proxy
 * 2. Start the collector (starts an HTTP server)
 * 3. Simulate some Applications with real Processes
 *
 */

require('./polyfill')

var spawn = require('child_process').spawn
var path = require('path')
var tmp = require('tmp')
var rimraf = require('rimraf')
var merge = require('merge')
var split = require('split')
var createDebug = require('debug')
var which = require('npm-which')(process.cwd())
var argv = require('yargs')
.usage(`Boots everything you need to try N|Solid Console.

Usage: $0 [options]`)
.describe('interval', 'delay (in milliseconds) between polling cycles')
.default('interval', 1000)
.nargs('interval', 1)
.describe('storage', 'location where snapshots and profiles may be stored')
.nargs('storage', 1)
.alias('s', 'storage')
.alias('h', 'help')
.help('h')
.argv

var port = process.env.NSOLID_CONSOLE_PORT || 3000

var storagePath, tmpobj

if (!argv.storage) {
  tmpobj = tmp.dirSync({ unsafeCleanup: true })

  storagePath = tmpobj.name
} else {
  storagePath = path.resolve(process.cwd(), argv.storage)
}

// Override `debug`'s default so buffers become strings
createDebug.coerce = function coerce (val) {
  if (val instanceof Error) return val.stack || val.message
  return val.toString()
}

process.on('uncaughtException', cleanupAndExit)

process.on('SIGINT', cleanupAndExit)
process.on('SIGTERM', cleanupAndExit)
process.on('beforeExit', cleanupAndExit)

function cleanupAndExit () {
  if (tmpobj) {
    try {
      rimraf.sync(tmpobj.name)
    } catch (e) {

    }
    tmpobj = null
  }
}

function debugWrap (proc, logger) {
  proc.stdout.pipe(process.stdout)
  proc.stderr.pipe(split()).on('data', logger)
  proc.on('close', logger.bind(null, 'closed with code'))
  return proc
}

var consoleEnv = merge(true, process.env, {})

if (argv['include-nsolid-console']) {
  consoleEnv.NSOLID_APPNAME = 'N|Solid'
  consoleEnv.NSOLID_SOCKET = ':0'
  consoleEnv.NSOLID_HUB = '127.0.0.1:4001'
}

debugWrap(spawn('etcd', ['--cors=*'], {}), createDebug('etcd'))

var pathToProxy = which.sync('nsolid-proxy')

debugWrap(spawn(pathToProxy, {
  cwd: process.cwd(),
  env: consoleEnv
}), createDebug('nsolid-proxy'))

debugWrap(spawn('node', [
  'bin/nsolid-console.js',
  '--interval', argv.interval || 1000,
  '--duration', 15,
  '--storage', storagePath,
  '--port', port
], {
  env: merge(true, process.env, consoleEnv, {
    env: {
      NODE_ENV: process.env.NODE_ENV || 'development'
    }
  })
}), createDebug('nsolid-console'))

debugWrap(spawn(
  'nsolid',
  ['bin/simulate.js', 'test/simulator/scenarios/noop.js@20'],
  {
    env: merge({
      'NSOLID_APPNAME': 'Shark Attack'
    }, process.env)
  }
), createDebug('scenario:SharkAttack'))

debugWrap(spawn(
  'nsolid',
  ['bin/simulate.js', 'test/simulator/scenarios/leaky-http.js@10'],
  {
    env: merge({
      'NSOLID_APPNAME': 'Beeeeeeeees'
    }, process.env)
  }
), createDebug('scenario:Beeeeeeeees'))
