#!/usr/bin/env node

/**
 * Production script for initialising N|Solid Console server
 * and starting polling of the N|Solid Hub.
 *
 * Does not concern itself with external dependencies such as etcd.
 */

process.title = 'nsolid-console'

require('./polyfill')

// hack around socket.io debugging BS
process.env.DEBUG = false

var fs = require('fs')
var os = require('os')
var path = require('path')
var rimraf = require('rimraf')

var collect = require('../src/collect')
var config = require('../src/collect/config')

var rc = require('rc')
var defaultTemporaryStorage = path.join(os.tmpdir(), 'nsolid-console')
var defaults = rc('nsolid-console', {
  hub: 'http://localhost:9000',
  timeout: 1000,
  storage: defaultTemporaryStorage,
  duration: 15,
  interval: 5000,
  port: process.env.PORT || 3000
})

var argv = require('yargs')
.usage(`Start the N|Solid Console server

Usage: $0 [options]
`.trim())
.describe('duration', 'Maximum time to retain graph data in minutes')
.default('duration', defaults.duration)
.alias('d', 'duration')
.nargs('duration', 1)
.describe('interface', 'Networking interface HTTP server will listen on')
.default('interface', '0.0.0.0')
.nargs('interface', 1)
.describe('interval', 'Frequency to poll N|Solid Hub in ms')
.default('interval', defaults.interval)
.alias('i', 'interval')
.nargs('interval', 1)
.describe('port', 'Port HTTP server will listen on')
.default('port', defaults.port)
.alias('p', 'port')
.nargs('port', 1)
.alias('r', 'record')
.describe('storage', 'Location on disk for storing snapshots & profiles')
.alias('s', 'storage')
.default('storage', defaults.storage)
.nargs('storage', 1)
.describe('hub', 'Location of the N|Solid Hub')
.default('hub', defaults.hub)
.nargs('hub', 1)
.describe('timeout', 'Maximum amount of time that a request to the N|Solid Hub can take in ms')
.default('timeout', defaults.timeout)
.nargs('timeout', 1)
.help('help')
.alias('help', 'h')
.argv

// Add the argv params to config
var conf = config(argv)

var storagePath = conf.storage

if (storagePath === defaultTemporaryStorage) {
  // clean storage dir if not supplied
  try {
    rimraf.sync(storagePath)
  } catch (err) {
    // ignore not found errs
    if (err.code !== 'ENOENT') onError(err)
  }
} else {
  storagePath = path.resolve(process.cwd(), storagePath)
}

var pollingEmitter = collect.poll(argv)

function writeToDisk (obj) {
  fs.appendFile(argv.record, JSON.stringify(obj) + '\n', function (err) {
    if (err) onError(err)
  })
}

if (argv.record) {
  pollingEmitter.on('output', writeToDisk)
}

collect.createServer(argv, pollingEmitter, function (err) {
  if (err) return onError(err)
  const listening = this.address()
  console.info('Server running on http://%s:%s', listening.address, listening.port)
}, onError)

function onError (err) {
  if (!err) return
  console.error(err.stack || err)

  // add helpful suggestions here
  if (err.code === 'EACCES') {
    console.error('\nHave you configured a writable --storage directory?')
  }

  console.error('\nRun `nsolid-console --help` for command-line help.')
  process.exit(1)
}
