/**
 * Server-side lib for collect
 */

var async = require('async')
var createRouter = require('routes')
var url = require('url')
var camelize = require('camelize')
var EventEmitter = require('events').EventEmitter
var path = require('path')
var fs = require('fs')
var mkdirp = require('mkdirp')
var spawn = require('child_process').spawn
var split = require('split')

var buildResponse = require('./collect/build-response')
var hub = require('./collect/hub')
var st = require('st')
var skateboard = require('skateboard')
var computeAppId = require('./compute-app-id')
var graphCache = require('./collect/graph-cache')
var createStorage = require('./collect/storage')
var defaultNodeEnv = 'production'

try {
  var stat = fs.statSync(path.join(__dirname, '..', 'webpack.config.js'))
  if (stat) {
    defaultNodeEnv = 'development'
  }
} catch (e) {}

// webpack dev/hot middleware configuration
var developmentMode = (process.env.NODE_ENV || defaultNodeEnv) !== 'production'

var commands = [
  'system_info', // hostname, totalmem, and metadata
  'system_stats',
  'process_stats'
]

// TODO: read these from disk
var snapshots = {}
var profiles = {}

function broadcastToHub (cmd, opts, cb) {
  if (typeof opts === 'function') {
    cb = opts
    opts = {}
  }

  hub.aggregate(cmd, opts, function broadcastComplete (e, o) {
    cb(e, o)
  })
}

function createHubStream (cmd, opts) {
  return hub.stream(cmd, opts)
}

function poll (options, broadcast) {
  var ee = new EventEmitter()

  var pollInterval = options.interval
  var duration = options.duration
  var storagePath = options.storage
  var graphDB = createStorage(storagePath, 'graphs')

  function run () {
    async.mapSeries(commands, broadcast || broadcastToHub, function (err, commandResults) {
      if (err) {
        setTimeout(run, pollInterval)
        return console.error(err)
      }

      var resultsByCommand = {}

      commands.forEach(function (name, i) {
        resultsByCommand[name] = clean(commandResults[i])
      })

      ee.emit('input', resultsByCommand)

      var processes = resultsByCommand['process_stats']
      var hosts = {}
      resultsByCommand['system_info'].forEach(function collectSystemInfo (system) {
        var host = system.reply

        if (!hosts[host.hostname]) {
          hosts[host.hostname] = host
        }
      })

      resultsByCommand['system_stats'].forEach(function collectSystemStats (system) {
        var hostname = system.meta.hostname
        var host = system.reply
        if (!hosts[hostname]) return

        if (!hosts[hostname].stats) {
          hosts[hostname].stats = host
          var totalmem = hosts[hostname].totalmem
          host.memoryPercentageUsed = ((totalmem - host.freemem) / totalmem) * 100
        }
      })

      var array = buildResponse(processes, hosts, pollInterval, duration)

      array.sort(function (a, b) {
        return a.name.localeCompare(b.name)
      })

      ee.emit('output', {
        hosts: hosts,
        applications: array,
        profiles: profiles,
        snapshots: snapshots
      })

      // for testing
      if (pollInterval !== false) {
        setTimeout(run, pollInterval)
      }
    })
  }

  // kick it off
  if (pollInterval !== 'never') {
    graphCache.init(graphDB, duration, run)
  }

  return ee
}

function clean (a) {
  // Until we can land a fix for https://github.com/nodesource/nsolid-epm-cli/issues/2
  if (!a) {
    return []
  }

  return a.filter(function (o) {
    // TODO: it's likely better to show the errors on the client if something
    //       bad is happening.  In all of the cases I've seen so far it's just
    //       an inability to resolve an ipv6 address (derp)
    return !!o.reply
  }).map(function (o) {
    o.reply = camelize(o.reply)
    return o
  })
}

function bailOnError (message, fn, context) {
  return function () {
    console.error(message)
    try {
      return fn.apply(context, arguments)
    } catch (e) {
      console.log(e.stack)
    }
  }
}

// TODO: refactor - this argument list is way too long
function respond (storagePath, storageDirectory, extension, res, params, cmd, store) {
  var c = hub.stream(cmd, {
    app: params.app,
    id: params.id
  })

  var time = Date.now()
  var filename = generateFilename(params.app, params.id, time, extension)

  c.pipe(fs.createWriteStream(path.join(storagePath, storageDirectory, filename)))

  c.on('error', function handleProfileError (err) {
    respondWithError(res, err)
  })

  c.on('end', function (code) {
    if (!store[params.app]) {
      store[params.app] = []
    }

    var insert = {
      time: time,
      app: params.app,
      id: params.id,
      date: new Date(),
      file: filename,
      title: params.title,
      pid: params.pid
    }

    // put the latest snapshots at the beginning of the array
    store[params.app].unshift(insert)

    res.setHeader('content-type', 'application/json')
    res.end(JSON.stringify(insert) + '\n')
  })
}

function generateFilename (appName, processId, time, extension) {
  var appId = computeAppId(appName)
  return [appId, processId, time].join('-') + '.' + extension
}

function respondWithError (res, err) {
  // TODO: proper error codes
  res.writeHead(404, { 'content-type': 'application/json' })
  res.end(JSON.stringify({
    message: err.message,
    code: err.code
  }) + '\n')
}

function respondFileExists (res, file) {
  fs.stat(file, function (e, r) {
    if (e) {
      res.writeHead(404)
    } else {
      res.writeHead(200, {
        'content-length': r.size
      })
    }

    res.end()
  })
}

function createServerRoutes (storagePath) {
  var router = createRouter()

  router.addRoute(
    '/heapsnapshot',
    function createHeapSnapshot (req, res, match) {
      var method = req.method.toLowerCase()
      var parsedUrl = url.parse(req.url, true)
      var params = parsedUrl.query
      var time = params.time || Date.now()
      var filename = generateFilename(
        params.app,
        params.id,
        time,
        'heapsnapshot'
      )
      var file = path.resolve(storagePath, path.join('snapshots', filename))

      if (method === 'head') {
        respondFileExists(res, file)
      } else if (method === 'get') {
        var readStream = fs.createReadStream(file)
        readStream.on('open', function downloadSnapshotOpen () {
          res.setHeader('Content-Type', 'application/json')
          res.setHeader('Content-Disposition', 'attachment; filename=' + filename)
          readStream.pipe(res)
        })

        readStream.on('error', function downloadSnapshotError (error) {
          respondWithError(res, error)
        })
      } else if (method === 'post') {
        var c = createHubStream('snapshot', params)

        c.pipe(fs.createWriteStream(file))

        c.on('error', function handleSnapshotError (err) {
          respondWithError(res, err)
        })

        c.on('end', function snapshotStreamEnd () {
          if (!snapshots[params.app]) {
            snapshots[params.app] = []
          }

          var insert = {
            time: time,
            app: params.app,
            id: params.id,
            date: new Date(),
            file: filename,
            title: params.title,
            pid: params.pid
          }

          // put the latest snapshots at the beginning of the array
          snapshots[params.app].unshift(insert)

          res.setHeader('content-type', 'application/json')
          res.end(JSON.stringify(insert) + '\n')
        })
      } else {
        res.end()
      }
    }
  )

  router.addRoute(
    '/cpuprofile',
    function createProfile (req, res, match) {
      var method = req.method.toLowerCase()
      var parsedUrl = url.parse(req.url, true)
      var params = parsedUrl.query
      var time = params.time || Date.now()
      var filename = generateFilename(
        params.app,
        params.id,
        time,
        'cpuprofile'
      )

      var file = path.resolve(storagePath, path.join('profiles', filename))

      if (method === 'head') {
        respondFileExists(res, file)
      } else if (method === 'get') {
        // e.g.: http://localhost:3000/applications/default-nsolid-scenario/processes/1062671c412debb5fdd538e08a333bd1f51dd152/profile?time=1438049368666
        var readStream = fs.createReadStream(file)
        readStream.on('open', function () {
          res.setHeader('Content-Type', 'application/json')
          res.setHeader('Content-Disposition', 'attachment; filename=' + filename)

          readStream.pipe(res)
        })

        readStream.on('error', function (error) {
          if (error.code === 'ENOENT') {
            res.writeHead(404)
          } else {
            res.writeHead(500)
          }
          res.end(error.toString())
        })
      } else if (method === 'post') {
        var duration = 5
        if (params.duration) {
          var intermediary = parseInt(params.duration, 10)
          if (!isNaN(intermediary)) {
            duration = intermediary
          }
        }

        // curl -X POST :3000/applications/default-nsolid-scenario/processes/1062671c412debb5fdd538e08a333bd1f51dd152/profile
        broadcastToHub('profile_start', {
          app: params.app,
          id: params.id
        }, function profileStarted (err) {
          if (err) return respondWithError(res, err)

          setTimeout(function () {
            respond(storagePath, 'profiles', 'cpuprofile', res, params, 'profile_stop', profiles)
          }, duration * 1000)
        })
      } else {
        res.end()
      }
    }
  )

  return router
}

function createServer (options, pollingEmitter, listening, error) {
  if (developmentMode) {
    var webpackConfig = require('../webpack.config')
    var webpack = require('webpack')
    var compiler = webpack(webpackConfig)
    var webpackDevMiddleware = require('webpack-dev-middleware')(compiler, webpackConfig.middleware)
    var webpackHotMiddleware = require('webpack-hot-middleware')(compiler, webpackConfig.hotMiddleware)
  }

  function processMiddleware (req, res, fn) {
    // only run the middleware when developmentMode is enabled
    if (developmentMode) {
      webpackDevMiddleware(req, res, function (err) {
        if (err) return fn(err)
        webpackHotMiddleware(req, res, fn)
      })
    } else {
      fn()
    }
  }

  var port = options.port
  var iface = options.interface
  var storagePath = options.storage

  var workerPath = path.join(__dirname, 'collect', 'heap-snapshot', 'worker.js')
  var router
  var applicationClients = []

  var appJsonPath = null

  if (storagePath) {
    var snapshotPath = path.join(storagePath, 'snapshots')
    var profilePath = path.join(storagePath, 'profiles')

    appJsonPath = path.join(storagePath, 'applications.json')

    try {
      mkdirp.sync(snapshotPath)
      mkdirp.sync(profilePath)
    } catch (err) {
      return error(err)
    }
    router = createServerRoutes(storagePath)
  }

  var mount = st({
    path: path.join(__dirname, '..', developmentMode ? 'assets' : 'build'),
    index: developmentMode ? false : 'index.html',
    passthrough: true
  })

  skateboard({
    port: port,
    interface: iface,
    dir: path.join(__dirname, '..', developmentMode ? 'assets' : 'build'),
    requestHandler: handleRequest,
    staticHandler: mount,
    listening: listening,
    error: error
  }, function (stream, params) {
    if (params.source === 'applications') {
      // keep track of connected clients
      applicationClients.push(stream)
      stream.on('end', function () {
        applicationClients = applicationClients.filter(function (a) {
          return a !== stream
        })
      })

      lastSend && stream.write(lastSend + '\n')
      stream.pipe(split()).on('data', function (line) {
        if (!line.trim().length) {
          return
        }

        var subscriptions = JSON.parse(line)
        // The only thing we care about right now is graph subscriptions
        graphCache.updateGraphSubscriptions(stream, subscriptions, function (e, d) {
          if (e || !d) {
            return
          }

          var graphs = graphCache.buildGraphPayload(stream)
          stream.write(JSON.stringify(Object.assign({}, lastOutput, graphs)) + '\n')
        })
      })
    } else if (params.source === 'heapsnapshot') {
      var proc = spawn('node', [workerPath, '--dir=' + snapshotPath], {
        stdio: 'pipe'
      })

      stream.once('end', bailOnError('client disconnected from heapsnapshot worker', proc.kill, proc))
      proc.on('exit', bailOnError('heapsnapshot worker exited', stream.close, stream))
      proc.stderr.pipe(process.stderr)
      stream.pipe(proc.stdin)

      proc.stdout.setEncoding('utf8')
      proc.stdout.pipe(stream)
    }
  })

  function handleRequest (req, res) {
    processMiddleware(req, res, function () {
      var urlPath = url.parse(req.url).pathname
      var match = router && router.match(urlPath)
      if (match) { // serve dynamic routes such as profiles and heap snapshots
        match.fn(req, res, match)
      } else { // serve static files
        mount(req, res, function () {
          req.url = '/'
          // `st` has cached the url as a 404, so override the cache
          req.sturl = '/'
          processMiddleware(req, res, function () {
            mount(req, res, function () {
              res.writeHead(404, 'Not Found')
              res.end('Not Found')
            })
          })
        })
      }
    })
  }

  var lastSend = null
  var lastOutput = null

  if (appJsonPath && exists(appJsonPath)) {
    lastSend = fs.readFileSync(appJsonPath, { encoding: 'utf-8' }) // this only runs once!
    try {
      lastOutput = JSON.parse(lastSend)
    } catch (e) {
      lastSend = null
      console.error('WARNING: could not load previous data')
    }
  }

  pollingEmitter.on('output', function (output) {
    lastOutput = output
    lastSend = JSON.stringify(output)
    if (appJsonPath) {
      fs.writeFile(appJsonPath, lastSend, function (err) {
        if (err) console.error(err.stack || err)
      })
    }

    applicationClients.forEach(function (client) {
      var graphs = graphCache.buildGraphPayload(client)
      client.write(JSON.stringify(Object.assign({}, output, graphs)) + '\n')
    })
  })
}

function exists (path) {
  try {
    return !!fs.statSync(path)
  } catch (e) {
    return !(e && e.code === 'ENOENT')
  }
}

// INTERFACE
module.exports.poll = poll
module.exports.createServer = createServer
